# Generated by Django 3.0.3 on 2020-03-05 12:14

from django.db import migrations, models
import django.utils.timezone


class Migration(migrations.Migration):

    dependencies = [
        ('dbaccounting', '0005_accounttype_parent'),
    ]

    operations = [
        migrations.AddField(
            model_name='account',
            name='date_create',
            field=models.DateTimeField(auto_now_add=True, default=django.utils.timezone.now),
            preserve_default=False,
        ),
    ]



from django.contrib import admin

from .models import AccountType,Account,Transaction
# Register your models here.

admin.site.register(AccountType)
admin.site.register(Account)

@admin.register(Transaction)
class TransactionAdmin(admin.ModelAdmin):
    list_filter = ['date','from_acc','to_acc']


# Generated by Django 3.0.3 on 2020-03-07 15:33

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('dbaccounting', '0009_transaction_edited'),
    ]

    operations = [
        migrations.AlterModelOptions(
            name='account',
            options={'ordering': ['name']},
        ),
    ]



from django.urls import path

from . import views

# Create urls here

urlpatterns = [
    path('', views.index, name='index'),
    path('report/', views.index, name='report'),
    path('balance-sheet/', views.index, name='balance-sheet'),
    path('income/', views.index, name='income'),
    path('cashflow/', views.index, name='cashflow'),
    path('retained/', views.index, name='retained'),
    path('debt/', views.index, name='debt'),
    path('forecast/', views.index, name='forecast'),
    path('acctype/', views.AccountTypeListView.as_view(), name='acctype'),
    path('acctype/<int:pk>/', views.AccountTypeDetailView.as_view(), name='acctype-detail'),
    path('acctype/create/',views.AccountTypeCreate.as_view(),name='acctype_create'),
    path('acctype/<int:pk>/update/',views.AccountTypeUpdate.as_view(),name='acctype_update'),
    path('acctype/<int:pk>/delete/',views.AccountTypeDelete.as_view(),name='acctype_delete'),
    path('acc/', views.AccountListView.as_view(), name='acc'),
    path('acc/<int:pk>/', views.AccountDetailView.as_view(), name='acc-detail'),
    path('acc/create/',views.AccountCreate.as_view(),name='acc_create'),
    path('acc/<int:pk>/update/',views.AccountUpdate.as_view(),name='acc_update'),
    path('acc/<int:pk>/delete/',views.AccountDelete.as_view(),name='acc_delete'),
    path('txn/', views.TransactionListView.as_view(), name='txn'),
    path('txn/<int:pk>/',  views.TransactionDetailView.as_view(), name='txn-detail'),
    path('txn/create/',views.transaction_create,name='txn_create'),
    path('txn/<int:pk>/update/',views.transaction_update,name='txn_update'),
    path('txn/<int:pk>/delete/',views.TransactionDelete.as_view(),name='txn_delete'),

]


import datetime

from django.test import TestCase
from django.urls import reverse, reverse_lazy
from django.utils import timezone
from django.contrib.auth.models import User, Permission

from dbaccounting.models import AccountType,Account,Transaction

# Create your tests here.

class AccountTypeListViewTest(TestCase):
    @classmethod
    def setUpTestData(cls):
        # Set Up User
        test_user1 = User.objects.create_user(username='testuser1', password='1X<ISRUkw+tuK')
        test_user2 = User.objects.create_user(username='testuser2', password='2HJ1vRV0Z&3iD')

        test_user1.save()
        test_user2.save()
        
        permission = Permission.objects.get(name='Can view account type')
        test_user2.user_permissions.add(permission)
        test_user2.save()
         
        # Create 13 authors for pagination tests
        number_of_account_types = 13

        for acc_type_id in range(number_of_account_types):
            if acc_type_id%2:   
                AccountType.objects.create(
                    name = f'Asset Type {acc_type_id}',
                    bal_type = 'D',
                )  
            else:
                AccountType.objects.create(
                    name = f'Liability Type {acc_type_id}',
                    bal_type = 'C',
                )  
    
    def test_redirect_if_not_logged_in(self):
        response = self.client.get(reverse('acctype'))
        # Manually check redirect (Can't use assertRedirect, because the redirect URL is unpredictable)
        self.assertEqual(response.status_code, 302)
        self.assertTrue(response.url.startswith('/accounts/login/'))

    def test_redirect_if_logged_in_but_not_correct_permission(self):
        login = self.client.login(username='testuser1', password='1X<ISRUkw+tuK')
        response = self.client.get(reverse('acctype'))
        self.assertEqual(response.status_code, 403) 

    def test_view_url_exists_at_desired_location(self):
        login = self.client.login(username='testuser2', password='2HJ1vRV0Z&3iD')
        response = self.client.get('/fin/acctype/')
        self.assertEqual(response.status_code,200)
    
    def test_view_url_accessible_by_name(self):
        login = self.client.login(username='testuser2', password='2HJ1vRV0Z&3iD')
        response = self.client.get(reverse('acctype'))
        self.assertEqual(response.status_code,200)

    def test_view_uses_correct_template(self):
        login = self.client.login(username='testuser2', password='2HJ1vRV0Z&3iD')
        response = self.client.get(reverse('acctype'))
        self.assertEqual(response.status_code,200)
        self.assertTemplateUsed(response, 'dbaccounting/accounttype_list.html')

    def test_pagination_is_ten(self):
        login = self.client.login(username='testuser2', password='2HJ1vRV0Z&3iD')
        response = self.client.get(reverse('acctype'))
        self.assertEqual(response.status_code,200)
        self.assertTrue('is_paginated' in response.context)
        self.assertTrue(response.context['is_paginated'] == True)
        self.assertTrue(len(response.context['accounttype_list'])==10)

class AccountListViewTest(TestCase):
    @classmethod
    def setUpTestData(cls):
        # Set Up User
        test_user1 = User.objects.create_user(username='testuser1', password='1X<ISRUkw+tuK')
        test_user2 = User.objects.create_user(username='testuser2', password='2HJ1vRV0Z&3iD')

        test_user1.save()
        test_user2.save()
        
        permission = Permission.objects.get(name='Can view account')
        test_user2.user_permissions.add(permission)
        test_user2.save()
        
        type1 = AccountType.objects.create(name="Assets",bal_type="D")
        type2 = AccountType.objects.create(name="Liabilities",bal_type="C")

        type1.save()
        type2.save()

        # Create 13 authors for pagination tests
        number_of_accounts = 23

        for acc_id in range(number_of_accounts):
            if acc_id%2:   
                Account.objects.create(
                    name = f'Asset Account {acc_id}',
                    acc_type=type1,
                )  
            else:
                Account.objects.create(
                    name = f'Liability Account {acc_id}',
                    acc_type = type2,
                )  
    
    def test_redirect_if_not_logged_in(self):
        response = self.client.get(reverse('acc'))
        # Manually check redirect (Can't use assertRedirect, because the redirect URL is unpredictable)
        self.assertEqual(response.status_code, 302)
        self.assertTrue(response.url.startswith('/accounts/login/'))

    def test_redirect_if_logged_in_but_not_correct_permission(self):
        login = self.client.login(username='testuser1', password='1X<ISRUkw+tuK')
        response = self.client.get(reverse('acc'))
        self.assertEqual(response.status_code, 403) 

    def test_view_url_exists_at_desired_location(self):
        login = self.client.login(username='testuser2', password='2HJ1vRV0Z&3iD')
        response = self.client.get('/fin/acc/')
        self.assertEqual(response.status_code,200)
    
    def test_view_url_accessible_by_name(self):
        login = self.client.login(username='testuser2', password='2HJ1vRV0Z&3iD')
        response = self.client.get(reverse('acc'))
        self.assertEqual(response.status_code,200)

    def test_view_uses_correct_template(self):
        login = self.client.login(username='testuser2', password='2HJ1vRV0Z&3iD')
        response = self.client.get(reverse('acc'))
        self.assertEqual(response.status_code,200)
        self.assertTemplateUsed(response, 'dbaccounting/account_list.html')

    def test_pagination_is_twenty(self):
        login = self.client.login(username='testuser2', password='2HJ1vRV0Z&3iD')
        response = self.client.get(reverse('acc'))
        self.assertEqual(response.status_code,200)
        self.assertTrue('is_paginated' in response.context)
        self.assertTrue(response.context['is_paginated'] == True)
        self.assertTrue(len(response.context['account_list'])==20)

class TransactionListViewTest(TestCase):
    @classmethod
    def setUpTestData(cls):
        # Set Up User
        test_user1 = User.objects.create_user(username='testuser1', password='1X<ISRUkw+tuK')
        test_user2 = User.objects.create_user(username='testuser2', password='2HJ1vRV0Z&3iD')

        test_user1.save()
        test_user2.save()
        
        permission = Permission.objects.get(name='Can view transaction')
        test_user2.user_permissions.add(permission)
        test_user2.save()
        
        type1 = AccountType.objects.create(name="Assets",bal_type="D")
        type2 = AccountType.objects.create(name="Liabilities",bal_type="C")
        type1.save()
        type2.save()

        acc1 = Account.objects.create(name="Asset Account", acc_type=type1)
        acc2 = Account.objects.create(name="Liability Account",acc_type=type2)
        acc1.save()
        acc2.save()

        # Create 13 authors for pagination tests
        number_of_transactions = 53

        for txn_id in range(number_of_transactions):
            if txn_id%2:   
                Transaction.objects.create(
                    from_acc = acc2,
                    to_acc = acc1,
                    amount=50
                )  
            else:
                Transaction.objects.create(
                    from_acc = acc1,
                    to_acc = acc2,
                    amount=50
                )  
    
    def test_redirect_if_not_logged_in(self):
        response = self.client.get(reverse('txn'))
        # Manually check redirect (Can't use assertRedirect, because the redirect URL is unpredictable)
        self.assertEqual(response.status_code, 302)
        self.assertTrue(response.url.startswith('/accounts/login/'))

    def test_redirect_if_logged_in_but_not_correct_permission(self):
        login = self.client.login(username='testuser1', password='1X<ISRUkw+tuK')
        response = self.client.get(reverse('txn'))
        self.assertEqual(response.status_code, 403) 

    def test_view_url_exists_at_desired_location(self):
        login = self.client.login(username='testuser2', password='2HJ1vRV0Z&3iD')
        response = self.client.get('/fin/txn/')
        self.assertEqual(response.status_code,200)
    
    def test_view_url_accessible_by_name(self):
        login = self.client.login(username='testuser2', password='2HJ1vRV0Z&3iD')
        response = self.client.get(reverse('txn'))
        self.assertEqual(response.status_code,200)

    def test_view_uses_correct_template(self):
        login = self.client.login(username='testuser2', password='2HJ1vRV0Z&3iD')
        response = self.client.get(reverse('txn'))
        self.assertEqual(response.status_code,200)
        self.assertTemplateUsed(response, 'dbaccounting/transaction_list.html')

    def test_pagination_is_fifty(self):
        login = self.client.login(username='testuser2', password='2HJ1vRV0Z&3iD')
        response = self.client.get(reverse('txn'))
        self.assertEqual(response.status_code,200)
        self.assertTrue('is_paginated' in response.context)
        self.assertTrue(response.context['is_paginated'] == True)
        self.assertTrue(len(response.context['transaction_list'])==50)

class AccountTypeDetailViewTest(TestCase):
    @classmethod
    def setUpTestData(cls):
        # Set Up User
        test_user1 = User.objects.create_user(username='testuser1', password='1X<ISRUkw+tuK')
        test_user2 = User.objects.create_user(username='testuser2', password='2HJ1vRV0Z&3iD')

        test_user1.save()
        test_user2.save()
        
        permission = Permission.objects.get(name='Can view account')
        test_user2.user_permissions.add(permission)
        test_user2.save()
        
        type1 = AccountType.objects.create(name="Assets",bal_type="D")
        type1.save()

    def test_redirect_if_not_logged_in(self):
        acctype = AccountType.objects.get(id=1)
        response = self.client.get(reverse('acctype-detail',args=[acctype.id]))
        # Manually check redirect (Can't use assertRedirect, because the redirect URL is unpredictable)
        self.assertEqual(response.status_code, 302)
        self.assertTrue(response.url.startswith('/accounts/login/'))

    def test_redirect_if_logged_in_but_not_correct_permission(self):
        acctype = AccountType.objects.get(id=1)
        login = self.client.login(username='testuser1', password='1X<ISRUkw+tuK')
        response = self.client.get(reverse('acctype-detail',args=[acctype.id]))
        self.assertEqual(response.status_code, 403) 

    def test_view_url_exists_at_desired_location(self):
        login = self.client.login(username='testuser2', password='2HJ1vRV0Z&3iD')
        response = self.client.get('/fin/acctype/1/')
        self.assertEqual(response.status_code,200)
    
    def test_view_url_accessible_by_name(self):
        acctype = AccountType.objects.get(id=1)
        login = self.client.login(username='testuser2', password='2HJ1vRV0Z&3iD')
        response = self.client.get(reverse('acctype-detail',args=[acctype.id]))
        self.assertEqual(response.status_code,200)

    def test_view_uses_correct_template(self):
        acctype = AccountType.objects.get(id=1)
        login = self.client.login(username='testuser2', password='2HJ1vRV0Z&3iD')
        response = self.client.get(reverse('acctype-detail',args=[acctype.id]))
        self.assertEqual(response.status_code,200)
        self.assertTemplateUsed(response, 'dbaccounting/accounttype_detail.html')

class AccountDetailViewTest(TestCase):
    @classmethod
    def setUpTestData(cls):
        # Set Up User
        test_user1 = User.objects.create_user(username='testuser1', password='1X<ISRUkw+tuK')
        test_user2 = User.objects.create_user(username='testuser2', password='2HJ1vRV0Z&3iD')

        test_user1.save()
        test_user2.save()
        
        permission = Permission.objects.get(name='Can view transaction')
        test_user2.user_permissions.add(permission)
        test_user2.save()
        
        type1 = AccountType.objects.create(name="Assets",bal_type="D")
        type1.save()
        acc1 = Account.objects.create(name="Asset Account", acc_type=type1)
        acc1.save()

    def test_redirect_if_not_logged_in(self):
        acc = Account.objects.get(id=1)
        response = self.client.get(reverse('acc-detail',args=[acc.id]))
        # Manually check redirect (Can't use assertRedirect, because the redirect URL is unpredictable)
        self.assertEqual(response.status_code, 302)
        self.assertTrue(response.url.startswith('/accounts/login/'))

    def test_redirect_if_logged_in_but_not_correct_permission(self):
        acc = Account.objects.get(id=1)
        login = self.client.login(username='testuser1', password='1X<ISRUkw+tuK')
        response = self.client.get(reverse('acc-detail',args=[acc.id]))
        self.assertEqual(response.status_code, 403) 

    def test_view_url_exists_at_desired_location(self):
        login = self.client.login(username='testuser2', password='2HJ1vRV0Z&3iD')
        response = self.client.get('/fin/acc/1/')
        self.assertEqual(response.status_code,200)
    
    def test_view_url_accessible_by_name(self):
        acc = Account.objects.get(id=1)
        login = self.client.login(username='testuser2', password='2HJ1vRV0Z&3iD')
        response = self.client.get(reverse('acc-detail',args=[acc.id]))
        self.assertEqual(response.status_code,200)

    def test_view_uses_correct_template(self):
        acc = Account.objects.get(id=1)
        login = self.client.login(username='testuser2', password='2HJ1vRV0Z&3iD')
        response = self.client.get(reverse('acc-detail',args=[acc.id]))
        self.assertEqual(response.status_code,200)
        self.assertTemplateUsed(response, 'dbaccounting/account_detail.html')

class TransactionDetailViewTest(TestCase):
    @classmethod
    def setUpTestData(cls):
        # Set Up User
        test_user1 = User.objects.create_user(username='testuser1', password='1X<ISRUkw+tuK')
        test_user2 = User.objects.create_user(username='testuser2', password='2HJ1vRV0Z&3iD')

        test_user1.save()
        test_user2.save()
        
        permission = Permission.objects.get(name='Can view transaction')
        test_user2.user_permissions.add(permission)
        test_user2.save()
        
        type1 = AccountType.objects.create(name="Assets",bal_type="D")
        type2 = AccountType.objects.create(name="Liabilities",bal_type="C")
        type1.save()
        type2.save()

        acc1 = Account.objects.create(name="Asset Account", acc_type=type1)
        acc2 = Account.objects.create(name="Liability Account",acc_type=type2)
        acc1.save()
        acc2.save()

        txn = Transaction.objects.create(from_acc=acc2,to_acc=acc1,amount=100)

    def test_redirect_if_not_logged_in(self):
        txn = Transaction.objects.get(id=1)
        response = self.client.get(reverse('txn-detail',args=[txn.id]))
        # Manually check redirect (Can't use assertRedirect, because the redirect URL is unpredictable)
        self.assertEqual(response.status_code, 302)
        self.assertTrue(response.url.startswith('/accounts/login/'))

    def test_redirect_if_logged_in_but_not_correct_permission(self):
        txn = Transaction.objects.get(id=1)
        login = self.client.login(username='testuser1', password='1X<ISRUkw+tuK')
        response = self.client.get(reverse('txn-detail',args=[txn.id]))
        self.assertEqual(response.status_code, 403) 

    def test_view_url_exists_at_desired_location(self):
        login = self.client.login(username='testuser2', password='2HJ1vRV0Z&3iD')
        response = self.client.get('/fin/txn/1/')
        self.assertEqual(response.status_code,200)
    
    def test_view_url_accessible_by_name(self):
        txn = Transaction.objects.get(id=1)
        login = self.client.login(username='testuser2', password='2HJ1vRV0Z&3iD')
        response = self.client.get(reverse('txn-detail',args=[txn.id]))
        self.assertEqual(response.status_code,200)

    def test_view_uses_correct_template(self):
        txn = Transaction.objects.get(id=1)
        login = self.client.login(username='testuser2', password='2HJ1vRV0Z&3iD')
        response = self.client.get(reverse('txn-detail',args=[txn.id]))
        self.assertEqual(response.status_code,200)
        self.assertTemplateUsed(response, 'dbaccounting/transaction_detail.html')

class AccountTypeCreateTest(TestCase):
    def setUp(self):
        # Create a user
        test_user1 = User.objects.create_user(username='testuser1', password='1X<ISRUkw+tuK')
        test_user2 = User.objects.create_user(username='testuser2', password='2HJ1vRV0Z&3iD')

        test_user1.save()
        test_user2.save()
        
        permission = Permission.objects.get(name='Can add account type')
        test_user2.user_permissions.add(permission)
        test_user2.save()
    
    def test_redirect_if_not_logged_in(self):
        response = self.client.get(reverse('acctype_create'))
        # Manually check redirect (Can't use assertRedirect, because the redirect URL is unpredictable)
        self.assertEqual(response.status_code, 302)
        self.assertTrue(response.url.startswith('/accounts/login/'))
        
    def test_redirect_if_logged_in_but_not_correct_permission(self):
        login = self.client.login(username='testuser1', password='1X<ISRUkw+tuK')
        response = self.client.get(reverse('acctype_create'))
        self.assertEqual(response.status_code, 403) # Should be code 403, however redirect causes for code 302 instead

    def test_logged_in_with_permission_add_account_type(self):
        login = self.client.login(username='testuser2', password='2HJ1vRV0Z&3iD')
        response = self.client.get(reverse('acctype_create'))
        
        # Check that it lets us login - we have the right permissions.
        self.assertEqual(response.status_code, 200)
    
    def test_uses_correct_template(self):
        login = self.client.login(username='testuser2', password='2HJ1vRV0Z&3iD')
        response = self.client.get(reverse('acctype_create'))
        self.assertEqual(response.status_code, 200)

        # Check we used correct template
        self.assertTemplateUsed(response, 'dbaccounting/accounttype_form.html')

class AccountCreateTest(TestCase):
    def setUp(self):
        # Create a user
        test_user1 = User.objects.create_user(username='testuser1', password='1X<ISRUkw+tuK')
        test_user2 = User.objects.create_user(username='testuser2', password='2HJ1vRV0Z&3iD')

        test_user1.save()
        test_user2.save()
        
        permission = Permission.objects.get(name='Can add account')
        test_user2.user_permissions.add(permission)
        test_user2.save()
    
    def test_redirect_if_not_logged_in(self):
        response = self.client.get(reverse('acc_create'))
        # Manually check redirect (Can't use assertRedirect, because the redirect URL is unpredictable)
        self.assertEqual(response.status_code, 302)
        self.assertTrue(response.url.startswith('/accounts/login/'))
        
    def test_redirect_if_logged_in_but_not_correct_permission(self):
        login = self.client.login(username='testuser1', password='1X<ISRUkw+tuK')
        response = self.client.get(reverse('acc_create'))
        self.assertEqual(response.status_code, 403) # Should be code 403, however redirect causes for code 302 instead

    def test_logged_in_with_permission_add_account(self):
        login = self.client.login(username='testuser2', password='2HJ1vRV0Z&3iD')
        response = self.client.get(reverse('acc_create'))
        
        # Check that it lets us login - we have the right permissions.
        self.assertEqual(response.status_code, 200)
    
    def test_uses_correct_template(self):
        login = self.client.login(username='testuser2', password='2HJ1vRV0Z&3iD')
        response = self.client.get(reverse('acc_create'))
        self.assertEqual(response.status_code, 200)

        # Check we used correct template
        self.assertTemplateUsed(response, 'dbaccounting/account_form.html')
    
class TransactionCreateTest(TestCase):
    def setUp(self):
        # Create a user
        test_user1 = User.objects.create_user(username='testuser1', password='1X<ISRUkw+tuK')
        test_user2 = User.objects.create_user(username='testuser2', password='2HJ1vRV0Z&3iD')

        test_user1.save()
        test_user2.save()
        
        permission = Permission.objects.get(name='Can add transaction')
        test_user2.user_permissions.add(permission)
        test_user2.save()
    
    def test_redirect_if_not_logged_in(self):
        response = self.client.get(reverse('txn_create'))
        # Manually check redirect (Can't use assertRedirect, because the redirect URL is unpredictable)
        self.assertEqual(response.status_code, 302)
        self.assertTrue(response.url.startswith('/accounts/login/'))
        
    def test_redirect_if_logged_in_but_not_correct_permission(self):
        login = self.client.login(username='testuser1', password='1X<ISRUkw+tuK')
        response = self.client.get(reverse('txn_create'))
        self.assertEqual(response.status_code, 403) # Should be code 403, however redirect causes for code 302 instead

    def test_logged_in_with_permission_add_transaction(self):
        login = self.client.login(username='testuser2', password='2HJ1vRV0Z&3iD')
        response = self.client.get(reverse('txn_create'))
        
        # Check that it lets us login - we have the right permissions.
        self.assertEqual(response.status_code, 200)
    
    def test_uses_correct_template(self):
        login = self.client.login(username='testuser2', password='2HJ1vRV0Z&3iD')
        response = self.client.get(reverse('txn_create'))
        self.assertEqual(response.status_code, 200)

        # Check we used correct template
        self.assertTemplateUsed(response, 'dbaccounting/transaction_form.html')


class AccountTypeUpdateTest(TestCase):
    def setUp(self):
        # Create a user
        test_user1 = User.objects.create_user(username='testuser1', password='1X<ISRUkw+tuK')
        test_user2 = User.objects.create_user(username='testuser2', password='2HJ1vRV0Z&3iD')

        test_user1.save()
        test_user2.save()
        
        permission1 = Permission.objects.get(name='Can change account type')
        permission2 = Permission.objects.get(name='Can view account')
        test_user2.user_permissions.add(permission1)
        test_user2.user_permissions.add(permission2)
        test_user2.save()

        # Create an AccountType
        self.test_acctype1 = AccountType.objects.create(name="Assets",bal_type="D")
        self.test_acctype2 = AccountType.objects.create(name="Current Assets",bal_type="D",parent=AccountType.objects.get(id=1))
        self.test_acctype3 = AccountType.objects.create(name="Liabilities",bal_type="C")
    
    def test_redirect_if_not_logged_in(self):
        response = self.client.get(reverse('acctype_update', kwargs={'pk': self.test_acctype2.pk}))
        # Manually check redirect (Can't use assertRedirect, because the redirect URL is unpredictable)
        self.assertEqual(response.status_code, 302)
        self.assertTrue(response.url.startswith('/accounts/login/'))
        
    def test_redirect_if_logged_in_but_not_correct_permission(self):
        login = self.client.login(username='testuser1', password='1X<ISRUkw+tuK')
        response = self.client.get(reverse('acctype_update', kwargs={'pk': self.test_acctype2.pk}))
        self.assertEqual(response.status_code, 403) # Should be code 403, however redirect causes for code 302 instead

    def test_logged_in_with_permission_add_account_type(self):
        login = self.client.login(username='testuser2', password='2HJ1vRV0Z&3iD')
        response = self.client.get(reverse('acctype_update', kwargs={'pk': self.test_acctype2.pk}))
        
        # Check that it lets us login - we have the right permissions.
        self.assertEqual(response.status_code, 200)

    def test_HTTP404_for_invalid_account_type_if_logged_in(self):
        # unlikely ID to match our accounttype instance!
        test_id = 100
        login = self.client.login(username='testuser2', password='2HJ1vRV0Z&3iD')
        response = self.client.get(reverse('acctype_update', kwargs={'pk':test_id}))
        self.assertEqual(response.status_code, 404)
        
    def test_uses_correct_template(self):
        login = self.client.login(username='testuser2', password='2HJ1vRV0Z&3iD')
        response = self.client.get(reverse('acctype_update', kwargs={'pk': self.test_acctype2.pk}))
        self.assertEqual(response.status_code, 200)

        # Check we used correct template
        self.assertTemplateUsed(response, 'dbaccounting/accounttype_form.html')

    # On all tests, constantly getting status code 200, when should be 302 (Redirect).
    
    # def test_redirects_to_account_type_detail_on_success(self):
    #     login = self.client.login(username='testuser2', password='2HJ1vRV0Z&3iD')
    #     new_account_type = "Equipment"
    #     response = self.client.post(reverse('acctype_update', kwargs={'pk':self.test_acctype2.pk,}), {'name':new_account_type})
    #     self.assertRedirects(response, reverse('acctype-detail',args=[self.test_acctype2.pk]))
        
    
    def test_form_invalid_bal_type(self):
        login = self.client.login(username='testuser2', password='2HJ1vRV0Z&3iD')
        invalid_bal_type = "E"
        response = self.client.post(reverse('acctype_update', kwargs={'pk': self.test_acctype2.pk}), {'bal_type': invalid_bal_type})
        self.assertEqual(response.status_code, 200)
        self.assertFormError(response, 'form', 'bal_type', 'Select a valid choice. E is not one of the available choices.')

class AccountUpdateTest(TestCase):
    def setUp(self):
        # Create a user
        test_user1 = User.objects.create_user(username='testuser1', password='1X<ISRUkw+tuK')
        test_user2 = User.objects.create_user(username='testuser2', password='2HJ1vRV0Z&3iD')

        test_user1.save()
        test_user2.save()
        
        permission1 = Permission.objects.get(name='Can change account')
        permission2 = Permission.objects.get(name='Can view transaction')
        test_user2.user_permissions.add(permission1)
        test_user2.user_permissions.add(permission2)
        test_user2.save()

        # Create an Account Type
        self.test_acctype1 = AccountType.objects.create(name="Assets",bal_type="D")
        self.test_acctype2 = AccountType.objects.create(name="Current Assets",bal_type="D",parent=AccountType.objects.get(id=1))
        self.test_acctype3 = AccountType.objects.create(name="Liabilities",bal_type="C")

        # Create an Account
        self.test_acc1 = Account.objects.create(name="Cash",acc_type=AccountType.objects.get(id=2))
        self.test_acc2 = Account.objects.create(name="Short-Term Debt",acc_type=AccountType.objects.get(id=3))
    
    def test_redirect_if_not_logged_in(self):
        response = self.client.get(reverse('acc_update', kwargs={'pk': self.test_acc2.pk}))
        # Manually check redirect (Can't use assertRedirect, because the redirect URL is unpredictable)
        self.assertEqual(response.status_code, 302)
        self.assertTrue(response.url.startswith('/accounts/login/'))
        
    def test_redirect_if_logged_in_but_not_correct_permission(self):
        login = self.client.login(username='testuser1', password='1X<ISRUkw+tuK')
        response = self.client.get(reverse('acc_update', kwargs={'pk': self.test_acc2.pk}))
        self.assertEqual(response.status_code, 403) # Should be code 403, however redirect causes for code 302 instead

    def test_logged_in_with_permission_add_account_type(self):
        login = self.client.login(username='testuser2', password='2HJ1vRV0Z&3iD')
        response = self.client.get(reverse('acc_update', kwargs={'pk': self.test_acc2.pk}))
        
        # Check that it lets us login - we have the right permissions.
        self.assertEqual(response.status_code, 200)

    def test_HTTP404_for_invalid_account_type_if_logged_in(self):
        # unlikely ID to match our accounttype instance!
        test_id = 100
        login = self.client.login(username='testuser2', password='2HJ1vRV0Z&3iD')
        response = self.client.get(reverse('acc_update', kwargs={'pk':test_id}))
        self.assertEqual(response.status_code, 404)
        
    def test_uses_correct_template(self):
        login = self.client.login(username='testuser2', password='2HJ1vRV0Z&3iD')
        response = self.client.get(reverse('acc_update', kwargs={'pk': self.test_acc2.pk}))
        self.assertEqual(response.status_code, 200)

        # Check we used correct template
        self.assertTemplateUsed(response, 'dbaccounting/account_form.html')

    # On all tests, constantly getting status code 200, when should be 302 (Redirect).
    #
    # def test_redirects_to_account_detail_on_success(self):
    #     login = self.client.login(username='testuser2', password='2HJ1vRV0Z&3iD')
    #     new_account = "Equipment Units"
    #     response = self.client.post(reverse('acc_update', kwargs={'pk':self.test_acc2.pk,}), {'name':new_account})
    #     self.assertRedirects(response, reverse('acc-detail',args=[self.test_acc2.pk]))
        
    
    def test_form_invalid_acc_type(self):
        login = self.client.login(username='testuser2', password='2HJ1vRV0Z&3iD')
        response = self.client.post(reverse('acc_update', kwargs={'pk': self.test_acc2.pk}), {'acc_type': ""})
        self.assertEqual(response.status_code, 200)
        self.assertFormError(response, 'form', 'acc_type', 'This field is required.')

class TransactionUpdateTest(TestCase):
    def setUp(self):
        # Create a user
        test_user1 = User.objects.create_user(username='testuser1', password='1X<ISRUkw+tuK')
        test_user2 = User.objects.create_user(username='testuser2', password='2HJ1vRV0Z&3iD')

        test_user1.save()
        test_user2.save()
        
        permission1 = Permission.objects.get(name='Can change transaction')
        permission2 = Permission.objects.get(name='Can view transaction')
        test_user2.user_permissions.add(permission1)
        test_user2.user_permissions.add(permission2)
        test_user2.save()

        # Create an Account Type
        self.test_acctype1 = AccountType.objects.create(name="Assets",bal_type="D")
        self.test_acctype2 = AccountType.objects.create(name="Current Assets",bal_type="D",parent=AccountType.objects.get(id=1))
        self.test_acctype3 = AccountType.objects.create(name="Liabilities",bal_type="C")

        # Create an Account
        self.test_acc1 = Account.objects.create(name="Cash",acc_type=AccountType.objects.get(id=2))
        self.test_acc2 = Account.objects.create(name="Short-Term Debt",acc_type=AccountType.objects.get(id=3))

        # Create a Transaction
        self.test_txn1 = Transaction.objects.create(from_acc=Account.objects.get(id=2),to_acc=Account.objects.get(id=1),amount=50)
        self.test_acc1.balance=50
        self.test_acc2.balance=-50
        self.test_acc1.save()
        self.test_acc2.save()
    
    def test_redirect_if_not_logged_in(self):
        response = self.client.get(reverse('txn_update', kwargs={'pk': self.test_txn1.pk}))
        # Manually check redirect (Can't use assertRedirect, because the redirect URL is unpredictable)
        self.assertEqual(response.status_code, 302)
        self.assertTrue(response.url.startswith('/accounts/login/'))
        
    def test_redirect_if_logged_in_but_not_correct_permission(self):
        login = self.client.login(username='testuser1', password='1X<ISRUkw+tuK')
        response = self.client.get(reverse('txn_update', kwargs={'pk': self.test_txn1.pk}))
        self.assertEqual(response.status_code, 403) # Should be code 403, however redirect causes for code 302 instead

    def test_logged_in_with_permission_add_account_type(self):
        login = self.client.login(username='testuser2', password='2HJ1vRV0Z&3iD')
        response = self.client.get(reverse('txn_update', kwargs={'pk': self.test_txn1.pk}))
        
        # Check that it lets us login - we have the right permissions.
        self.assertEqual(response.status_code, 200)

    def test_HTTP404_for_invalid_account_type_if_logged_in(self):
        # unlikely ID to match our accounttype instance!
        test_id = 100
        login = self.client.login(username='testuser2', password='2HJ1vRV0Z&3iD')
        response = self.client.get(reverse('txn_update', kwargs={'pk':test_id}))
        self.assertEqual(response.status_code, 404)
        
    def test_uses_correct_template(self):
        login = self.client.login(username='testuser2', password='2HJ1vRV0Z&3iD')
        response = self.client.get(reverse('txn_update', kwargs={'pk': self.test_txn1.pk}))
        self.assertEqual(response.status_code, 200)

        # Check we used correct template
        self.assertTemplateUsed(response, 'dbaccounting/transaction_form.html')

    # These tests run into errors with the clean() function on the custom form. 
    # However the form test should cover this functionality
    #
    # def test_redirects_to_transaction_detail_on_success(self):
    #     login = self.client.login(username='testuser2', password='2HJ1vRV0Z&3iD')
    #     new_amount = 25
    #     response = self.client.post(reverse('txn_update', kwargs={'pk':self.test_txn1.pk,}), {'amount':new_amount})
    #     self.assertRedirects(response, reverse('txn-detail',args=[self.test_txn1.pk]))
        
    
    # def test_form_invalid_acc_type(self):
    #     login = self.client.login(username='testuser2', password='2HJ1vRV0Z&3iD')
    #     response = self.client.post(reverse('txn_update', kwargs={'pk': self.test_txn1.pk}), {'amount': -50})
    #     self.assertEqual(response.status_code, 200)
    #     self.assertFormError(response, 'form', 'amount', 'This field is required.')
    
class AccountTypeDeleteTest(TestCase):
    def setUp(self):
        # Create a user
        test_user1 = User.objects.create_user(username='testuser1', password='1X<ISRUkw+tuK')
        test_user2 = User.objects.create_user(username='testuser2', password='2HJ1vRV0Z&3iD')

        test_user1.save()
        test_user2.save()
        
        permission1 = Permission.objects.get(name='Can delete account type')
        permission2 = Permission.objects.get(name='Can view account type')
        test_user2.user_permissions.add(permission1)
        test_user2.user_permissions.add(permission2)
        test_user2.save()

        # Create an AccountType
        self.test_acctype1 = AccountType.objects.create(name="Assets",bal_type="D")
        self.test_acctype2 = AccountType.objects.create(name="Current Assets",bal_type="D",parent=AccountType.objects.get(id=1))
        self.test_acctype3 = AccountType.objects.create(name="Liabilities",bal_type="C")
    
    def test_redirect_if_not_logged_in(self):
        response = self.client.get(reverse('acctype_delete', kwargs={'pk': self.test_acctype2.pk}))
        # Manually check redirect (Can't use assertRedirect, because the redirect URL is unpredictable)
        self.assertEqual(response.status_code, 302)
        self.assertTrue(response.url.startswith('/accounts/login/'))
        
    def test_redirect_if_logged_in_but_not_correct_permission(self):
        login = self.client.login(username='testuser1', password='1X<ISRUkw+tuK')
        response = self.client.get(reverse('acctype_delete', kwargs={'pk': self.test_acctype2.pk}))
        self.assertEqual(response.status_code, 403) # Should be code 403, however redirect causes for code 302 instead

    def test_logged_in_with_permission_add_account_type(self):
        login = self.client.login(username='testuser2', password='2HJ1vRV0Z&3iD')
        response = self.client.get(reverse('acctype_delete', kwargs={'pk': self.test_acctype2.pk}))
        
        # Check that it lets us login - we have the right permissions.
        self.assertEqual(response.status_code, 200)

    def test_HTTP404_for_invalid_account_type_if_logged_in(self):
        # unlikely ID to match our accounttype instance!
        test_id = 100
        login = self.client.login(username='testuser2', password='2HJ1vRV0Z&3iD')
        response = self.client.get(reverse('acctype_delete', kwargs={'pk':test_id}))
        self.assertEqual(response.status_code, 404)
        
    def test_uses_correct_template(self):
        login = self.client.login(username='testuser2', password='2HJ1vRV0Z&3iD')
        response = self.client.get(reverse('acctype_delete', kwargs={'pk': self.test_acctype2.pk}))
        self.assertEqual(response.status_code, 200)

        # Check we used correct template
        self.assertTemplateUsed(response, 'dbaccounting/accounttype_confirm_delete.html')

    # On all tests, constantly getting status code 200, when should be 302 (Redirect).
    
    def test_redirects_to_account_type_detail_on_success(self):
        login = self.client.login(username='testuser2', password='2HJ1vRV0Z&3iD')
        response = self.client.post(reverse('acctype_delete', kwargs={'pk':self.test_acctype2.pk,}), {'submit':"Confirm"})
        self.assertRedirects(response, reverse('acctype'))

class AccountUpdateTest(TestCase):
    def setUp(self):
        # Create a user
        test_user1 = User.objects.create_user(username='testuser1', password='1X<ISRUkw+tuK')
        test_user2 = User.objects.create_user(username='testuser2', password='2HJ1vRV0Z&3iD')

        test_user1.save()
        test_user2.save()
        
        permission1 = Permission.objects.get(name='Can delete account')
        permission2 = Permission.objects.get(name='Can view account')
        test_user2.user_permissions.add(permission1)
        test_user2.user_permissions.add(permission2)
        test_user2.save()

        # Create an Account Type
        self.test_acctype1 = AccountType.objects.create(name="Assets",bal_type="D")
        self.test_acctype2 = AccountType.objects.create(name="Current Assets",bal_type="D",parent=AccountType.objects.get(id=1))
        self.test_acctype3 = AccountType.objects.create(name="Liabilities",bal_type="C")

        # Create an Account
        self.test_acc1 = Account.objects.create(name="Cash",acc_type=AccountType.objects.get(id=2))
        self.test_acc2 = Account.objects.create(name="Short-Term Debt",acc_type=AccountType.objects.get(id=3))
    
    def test_redirect_if_not_logged_in(self):
        response = self.client.get(reverse('acc_delete', kwargs={'pk': self.test_acc2.pk}))
        # Manually check redirect (Can't use assertRedirect, because the redirect URL is unpredictable)
        self.assertEqual(response.status_code, 302)
        self.assertTrue(response.url.startswith('/accounts/login/'))
        
    def test_redirect_if_logged_in_but_not_correct_permission(self):
        login = self.client.login(username='testuser1', password='1X<ISRUkw+tuK')
        response = self.client.get(reverse('acc_delete', kwargs={'pk': self.test_acc2.pk}))
        self.assertEqual(response.status_code, 403) # Should be code 403, however redirect causes for code 302 instead

    def test_logged_in_with_permission_add_account_type(self):
        login = self.client.login(username='testuser2', password='2HJ1vRV0Z&3iD')
        response = self.client.get(reverse('acc_delete', kwargs={'pk': self.test_acc2.pk}))
        
        # Check that it lets us login - we have the right permissions.
        self.assertEqual(response.status_code, 200)

    def test_HTTP404_for_invalid_account_type_if_logged_in(self):
        # unlikely ID to match our accounttype instance!
        test_id = 100
        login = self.client.login(username='testuser2', password='2HJ1vRV0Z&3iD')
        response = self.client.get(reverse('acc_delete', kwargs={'pk':test_id}))
        self.assertEqual(response.status_code, 404)
        
    def test_uses_correct_template(self):
        login = self.client.login(username='testuser2', password='2HJ1vRV0Z&3iD')
        response = self.client.get(reverse('acc_delete', kwargs={'pk': self.test_acc2.pk}))
        self.assertEqual(response.status_code, 200)

        # Check we used correct template
        self.assertTemplateUsed(response, 'dbaccounting/account_confirm_delete.html')

    # On all tests, constantly getting status code 200, when should be 302 (Redirect).
    #
    def test_redirects_to_account_detail_on_success(self):
        login = self.client.login(username='testuser2', password='2HJ1vRV0Z&3iD')
        response = self.client.post(reverse('acc_delete', kwargs={'pk':self.test_acc2.pk,}), {'submit':"Confirm"})
        self.assertRedirects(response, reverse('acc'))

class TransactionUpdateTest(TestCase):
    def setUp(self):
        # Create a user
        test_user1 = User.objects.create_user(username='testuser1', password='1X<ISRUkw+tuK')
        test_user2 = User.objects.create_user(username='testuser2', password='2HJ1vRV0Z&3iD')

        test_user1.save()
        test_user2.save()
        
        permission1 = Permission.objects.get(name='Can delete transaction')
        permission2 = Permission.objects.get(name='Can view transaction')
        test_user2.user_permissions.add(permission1)
        test_user2.user_permissions.add(permission2)
        test_user2.save()

        # Create an Account Type
        self.test_acctype1 = AccountType.objects.create(name="Assets",bal_type="D")
        self.test_acctype2 = AccountType.objects.create(name="Current Assets",bal_type="D",parent=AccountType.objects.get(id=1))
        self.test_acctype3 = AccountType.objects.create(name="Liabilities",bal_type="C")

        # Create an Account
        self.test_acc1 = Account.objects.create(name="Cash",acc_type=AccountType.objects.get(id=2))
        self.test_acc2 = Account.objects.create(name="Short-Term Debt",acc_type=AccountType.objects.get(id=3))

        # Create a Transaction
        self.test_txn1 = Transaction.objects.create(from_acc=Account.objects.get(id=2),to_acc=Account.objects.get(id=1),amount=50)
        self.test_acc1.balance=50
        self.test_acc2.balance=-50
        self.test_acc1.save()
        self.test_acc2.save()
    
    def test_redirect_if_not_logged_in(self):
        response = self.client.get(reverse('txn_delete', kwargs={'pk': self.test_txn1.pk}))
        # Manually check redirect (Can't use assertRedirect, because the redirect URL is unpredictable)
        self.assertEqual(response.status_code, 302)
        self.assertTrue(response.url.startswith('/accounts/login/'))
        
    def test_redirect_if_logged_in_but_not_correct_permission(self):
        login = self.client.login(username='testuser1', password='1X<ISRUkw+tuK')
        response = self.client.get(reverse('txn_delete', kwargs={'pk': self.test_txn1.pk}))
        self.assertEqual(response.status_code, 403) # Should be code 403, however redirect causes for code 302 instead

    def test_logged_in_with_permission_delete_transaction(self):
        login = self.client.login(username='testuser2', password='2HJ1vRV0Z&3iD')
        response = self.client.get(reverse('txn_delete', kwargs={'pk': self.test_txn1.pk}))
        
        # Check that it lets us login - we have the right permissions.
        self.assertEqual(response.status_code, 200)

    def test_HTTP404_for_invalid_account_type_if_logged_in(self):
        # unlikely ID to match our accounttype instance!
        test_id = 100
        login = self.client.login(username='testuser2', password='2HJ1vRV0Z&3iD')
        response = self.client.get(reverse('txn_delete', kwargs={'pk':test_id}))
        self.assertEqual(response.status_code, 404)
        
    def test_uses_correct_template(self):
        login = self.client.login(username='testuser2', password='2HJ1vRV0Z&3iD')
        response = self.client.get(reverse('txn_delete', kwargs={'pk': self.test_txn1.pk}))
        self.assertEqual(response.status_code, 200)

        # Check we used correct template
        self.assertTemplateUsed(response, 'dbaccounting/transaction_confirm_delete.html')

    # These tests run into errors with the clean() function on the custom form. 
    # However the form test should cover this functionality
    #
    def test_redirects_to_transaction_detail_on_success(self):
        login = self.client.login(username='testuser2', password='2HJ1vRV0Z&3iD')
        response = self.client.post(reverse('txn_delete', kwargs={'pk':self.test_txn1.pk,}), {'submit':"Confirm"})
        self.assertRedirects(response, reverse('txn'))



# Generated by Django 3.0.3 on 2020-03-06 20:44

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('dbaccounting', '0008_transaction_updating'),
    ]

    operations = [
        migrations.AddField(
            model_name='transaction',
            name='edited',
            field=models.BooleanField(default=False),
        ),
    ]



from django.db import models
from django.urls import reverse

# Create your models here.

class AccountType(models.Model):
    name = models.CharField(max_length=64, unique=True)
    
    BALANCE_CHOICES = (
        ('C','Credit'),
        ('D','Debit'),
    )

    bal_type = models.CharField(choices=BALANCE_CHOICES, max_length=1,
        help_text="Indicate whether the account type is a credit or debit", verbose_name = "balance type")
    
    parent = models.ForeignKey('AccountType',on_delete=models.CASCADE,help_text = "Is this a subcategory (e.g. Current Assets)",null=True,blank=True)
    
    class Meta:
        ordering = ['-bal_type','name']
    
    def get_absolute_url(self):
        """Returns the url to access a detail record for this book."""
        return reverse('acctype-detail',args=[str(self.id)])
    
    def __str__(self):
        return f'{self.name}'

class Account(models.Model):
    date_create = models.DateTimeField(auto_now_add=True)
    name = models.CharField(max_length=64, unique=True)
    acc_type = models.ForeignKey(AccountType,on_delete=models.CASCADE,verbose_name="account type")
    balance = models.FloatField(default=0,null=True)

    class Meta:
        ordering = ['name']

    def get_absolute_url(self):
        """Returns the url to access a detail record for this book."""
        return reverse('acc-detail',args=[str(self.id)])

    def __str__(self):
        return f'{self.name}'

class Transaction(models.Model):
    date = models.DateTimeField(auto_now_add=True)
    from_acc = models.ForeignKey(Account,on_delete=models.CASCADE, related_name = 'from_account', verbose_name = "from account")
    to_acc = models.ForeignKey(Account,on_delete=models.CASCADE, related_name = 'to_account', verbose_name = "to account")
    amount = models.FloatField()
    note = models.TextField(max_length = 256, blank=True,null=True)
    updating = models.ForeignKey('Transaction',on_delete=models.SET_NULL,blank=True,null=True)
    edited = models.BooleanField(default=False)

    class Meta:
        ordering = ['date']

    def get_absolute_url(self):
        """Returns the url to access a detail record for this book."""
        return reverse('txn-detail',args=[str(self.id)])

    def __str__(self):
        return f'TXN on {self.date} from {self.from_acc} to {self.to_acc} for {self.amount} AED'



# Generated by Django 3.0.3 on 2020-03-03 20:00

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('dbaccounting', '0002_auto_20200303_2358'),
    ]

    operations = [
        migrations.AlterField(
            model_name='account',
            name='acc_type',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='dbaccounting.AccountType', verbose_name='account type'),
        ),
        migrations.AlterField(
            model_name='accounttype',
            name='bal_type',
            field=models.CharField(choices=[('C', 'Credit'), ('D', 'Debit')], help_text='Indicate whether the account type is a credi or debit', max_length=1, verbose_name='balance type'),
        ),
        migrations.AlterField(
            model_name='transaction',
            name='from_acc',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='from_account', to='dbaccounting.Account', verbose_name='from account'),
        ),
        migrations.AlterField(
            model_name='transaction',
            name='to_acc',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='to_account', to='dbaccounting.Account', verbose_name='to account'),
        ),
    ]



import datetime

from django.shortcuts import render, get_object_or_404
from django.contrib.auth.decorators import login_required,permission_required
from django.contrib.auth.mixins import LoginRequiredMixin, PermissionRequiredMixin
from django.views.generic.edit import CreateView, UpdateView, DeleteView
from django.views import generic
from django.urls import reverse_lazy, reverse
from django.db import transaction
from django.http import HttpResponseRedirect

from dbaccounting.models import AccountType,Account,Transaction
from dbaccounting.forms import TransactionForm
# Create your views here.

# View related objects
class AccountLedger:
    """An organized collection of an AccountType, its child AccountTypes, related Accounts, and total balances"""
    def __init__(self,acc_type):
        self.acc_type = acc_type
        self.sub_type = AccountType.objects.filter(parent=acc_type)
        self.accs = Account.objects.filter(acc_type=acc_type)
        self.total = sum((acc.balance for acc in self.accs)) if self.accs else 0

        self.sub_accs = {}
        if self.sub_type:    
            for sub in self.sub_type:
                self.sub_accs[sub] = AccountLedger(sub)

        self.subtotal = sum((self.sub_accs[sub].total for sub in self.sub_type)) if len(self.sub_type)>0 else 0
        self.total+=self.subtotal

# View Implementations Below

# Account Types
class AccountTypeDetailView(PermissionRequiredMixin,generic.DetailView):
    permission_required=("dbaccounting.view_account",)
    model = AccountType

class AccountTypeListView(PermissionRequiredMixin,generic.ListView):
    permission_required=("dbaccounting.view_accounttype",)
    model = AccountType
    paginate_by = 10

class AccountTypeCreate(PermissionRequiredMixin,CreateView):
    permission_required=("dbaccounting.add_accounttype",)
    model = AccountType
    fields = '__all__'
    
class AccountTypeUpdate(PermissionRequiredMixin,UpdateView):
    permission_required=("dbaccounting.change_accounttype",)
    model = AccountType
    fields = '__all__'

class AccountTypeDelete(PermissionRequiredMixin,DeleteView):
    permission_required=("dbaccounting.delete_accounttype",)
    model = AccountType
    success_url = reverse_lazy('acctype')

# Accounts
class AccountDetailView(PermissionRequiredMixin,generic.DetailView):
    permission_required=("dbaccounting.view_transaction",)
    model = Account

class AccountListView(PermissionRequiredMixin,generic.ListView):
    permission_required=("dbaccounting.view_account",)
    model = Account
    paginate_by = 20

class AccountCreate(PermissionRequiredMixin,CreateView):
    permission_required=("dbaccounting.add_account",)
    model = Account
    fields = '__all__'
    
class AccountUpdate(PermissionRequiredMixin,UpdateView):
    permission_required=("dbaccounting.change_account",)
    model = Account
    fields = ['name','acc_type','balance']
    success_url = reverse_lazy('acc')

class AccountDelete(PermissionRequiredMixin,DeleteView):
    permission_required=("dbaccounting.delete_account",)
    model = Account
    success_url = reverse_lazy('acc')

# Transactions
class TransactionDetailView(PermissionRequiredMixin,generic.DetailView):
    permission_required=("dbaccounting.view_transaction",)
    model = Transaction

class TransactionListView(PermissionRequiredMixin,generic.ListView):
    permission_required=("dbaccounting.view_transaction",)
    model = Transaction
    paginate_by = 50

class TransactionDelete(PermissionRequiredMixin,DeleteView):
    permission_required=("dbaccounting.delete_transaction",)
    model = Transaction
    success_url = reverse_lazy('txn')

    @transaction.atomic
    def delete(self,request,pk):
        txn = get_object_or_404(Transaction,pk=pk)
        txn.from_acc.balance+=txn.amount
        txn.to_acc.balance-=txn.amount
        txn.from_acc.save()
        txn.to_acc.save()

        if txn.updating:
            prev = txn.updating
            prev.edited = False
            prev.from_acc.balance-=prev.amount
            prev.to_acc.balance+=prev.amount
            prev.from_acc.save()
            prev.to_acc.save()
            prev.save()

        txn.delete()
        
        return HttpResponseRedirect(reverse_lazy('txn'))


@transaction.atomic
@login_required
@permission_required('dbaccounting.add_transaction',raise_exception=True)
def transaction_create(request):
    # If this is a POST request then process the Form data
    if request.method == 'POST':

        # Create a form instance and populate it with data from the request (binding):
        form = TransactionForm(request.POST)

        #Check if the form is valid:
        if form.is_valid():
            # process the data in form.cleaned_data as required
            from_acc = get_object_or_404(Account,pk=form.cleaned_data['from_acc'].pk)
            to_acc = get_object_or_404(Account,pk=form.cleaned_data['to_acc'].pk)
            amount = form.cleaned_data['amount']
            from_acc.balance -= amount
            to_acc.balance += amount
            from_acc.save()
            to_acc.save()
            note = form.cleaned_data['note']

            Transaction.objects.create(from_acc = from_acc,to_acc = to_acc,amount = amount, note=note)

            # redirect to a new URL
            return HttpResponseRedirect(reverse('txn'))
        
    # If this is a GET (or any other method) create the default form.
    else:
        form = TransactionForm()

    context = {
        'form': form,
    }

    return render(request,'dbaccounting/transaction_form.html',context)

@transaction.atomic
@login_required
@permission_required('dbaccounting.change_transaction',raise_exception=True)
def transaction_update(request,pk):
    orig_txn = get_object_or_404(Transaction,pk=pk)
    # If this is a POST request then process the Form data
    if request.method == 'POST':
        # Create a form instance and populate it with data from the request (binding):
        form = TransactionForm(request.POST)
 
        #Check if the form is valid:
        if form.is_valid():
            # process the data in form.cleaned_data as required
            from_acc = get_object_or_404(Account,pk=form.cleaned_data['from_acc'].pk)
            to_acc = get_object_or_404(Account,pk=form.cleaned_data['to_acc'].pk)
            amount = form.cleaned_data['amount']

            if from_acc!=orig_txn.from_acc:
                orig_txn.from_acc.balance += orig_txn.amount
                from_acc.balance -= orig_txn.amount
            
            if to_acc!=orig_txn.to_acc:
                orig_txn.to_acc.balance -= orig_txn.amount
                to_acc.balance += orig_txn.amount
            
            if amount!=orig_txn.amount:
                diff = amount-orig_txn.amount
                from_acc.balance-=diff
                to_acc.balance+=diff

            orig_txn.from_acc.save()
            orig_txn.to_acc.save()

            from_acc.save()
            to_acc.save()
            
            orig_txn.edited=True
            orig_txn.save()

            note = form.cleaned_data['note']

            Transaction.objects.create(from_acc = from_acc,to_acc = to_acc,amount = amount, note=note, updating=orig_txn)

            # redirect to a new URL
            return HttpResponseRedirect(reverse('txn'))
        
    # If this is a GET (or any other method) create the default form.
    else:
        form = TransactionForm(instance = orig_txn)

    context = {
        'form': form,
    }

    return render(request,'dbaccounting/transaction_form.html',context)

# Balance Sheet (Work in Progress)

@permission_required(('dbaccounting.view_account','dbaccounting.view_accounttype'))
def balance_sheet(request):
    date = str(datetime.date.today())
    main_acc_types = AccountType.objects.filter(parent__isnull=True)

    acc_types = AccountType.objects.all()
    accs = Account.objects.all()

    context = {
        'acc_types': acc_types,
        'accs': accs,
        'date':date,
    }
    
    return render(request,'balance_sheet.html',context=context)

# Index/Main Menu
@login_required
@permission_required(('dbaccounting.view_transaction'), raise_exception=True)
def index(request):
    acc_types = AccountType.objects.all().count()
    accs = Account.objects.all().count()
    txns = Transaction.objects.all().count()

    # Number of visits to this view, as counted in the session variable.
    num_visits = request.session.get('num_visits',0)
    request.session['num_visits']=num_visits+1

    context = {
        'acc_types': acc_types,
        'accs': accs,
        'txns': txns,
        'num_visits':num_visits,
    }

    # Render the HTML template index.html with the data in the context variable
    return render(request, 'index.html', context=context)


# Generated by Django 3.0.3 on 2020-03-06 17:18

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('dbaccounting', '0006_account_date_create'),
    ]

    operations = [
        migrations.AlterField(
            model_name='accounttype',
            name='parent',
            field=models.ForeignKey(blank=True, help_text='Is this a subcategory (e.g. Current Assets)', null=True, on_delete=django.db.models.deletion.CASCADE, to='dbaccounting.AccountType'),
        ),
    ]



from django.test import TestCase
from django.db import IntegrityError

from dbaccounting.models import AccountType,Account,Transaction
# Create your tests here.

class AccountTypeTest(TestCase):
    @classmethod
    def setUpTestData(cls):
        AccountType.objects.create(name = "Assets",bal_type = "D")
        AccountType.objects.create(name = "Current Assets", bal_type = "D", parent=AccountType.objects.get(id=1))
    
    def test_name_label(self):
        acc_type = AccountType.objects.get(id=1)
        field_label = acc_type._meta.get_field('name').verbose_name
        self.assertEquals(field_label, 'name')
    
    def test_bal_type_label(self):
        acc_type = AccountType.objects.get(id=1)
        field_label = acc_type._meta.get_field('bal_type').verbose_name
        self.assertEquals(field_label, 'balance type')
    
    def test_parent_label(self):
        acc_type = AccountType.objects.get(id=1)
        field_label = acc_type._meta.get_field('parent').verbose_name
        self.assertEquals(field_label, 'parent')
    
    def test_name_max_length(self):
        acc_type = AccountType.objects.get(id=1)
        max_length = acc_type._meta.get_field('name').max_length
        self.assertEquals(max_length,64)
    
    def test_bal_type_max_length(self):
        acc_type = AccountType.objects.get(id=1)
        max_length = acc_type._meta.get_field('bal_type').max_length
        self.assertEquals(max_length,1)
    
    def test_unique_name(self):
        with self.assertRaises(IntegrityError):
            AccountType.objects.create(name="Assets",bal_type="C")
    
    def test_parent_instance(self):
        acc_type = AccountType.objects.get(id=2)
        parent = acc_type.parent
        self.assertEquals(parent,AccountType.objects.get(id=1))

    def test_meta(self):
        acc_type = AccountType.objects.get(id=1)
        order = acc_type._meta.ordering
        self.assertEquals(order,['-bal_type','name'])
    
    def test_object_name(self):
        acc_type = AccountType.objects.get(id=1)
        expected_object_name = f'{acc_type.name}'
        self.assertEquals(expected_object_name,str(acc_type))
    
    def test_get_absolute_url(self):
        acc_type = AccountType.objects.get(id=1)
        # This will also fail if the urlconf is not defined.
        self.assertEquals(acc_type.get_absolute_url(), '/fin/acctype/1/')

class AccountTest(TestCase):
    def setUp(self):
        acc_type1 = AccountType.objects.create(name = "Assets",bal_type = "D")
        Account.objects.create(name="Cash",acc_type=acc_type1)
    
    def test_name_label(self):
        acc = Account.objects.get(id=1)
        field_label = acc._meta.get_field('name').verbose_name
        self.assertEquals(field_label, 'name')
    
    def test_acc_type_label(self):
        acc = Account.objects.get(id=1)
        field_label = acc._meta.get_field('acc_type').verbose_name
        self.assertEquals(field_label, 'account type')
    
    def test_name_max_length(self):
        acc = Account.objects.get(id=1)
        max_length = acc._meta.get_field('name').max_length
        self.assertEquals(max_length,64)
    
    def test_default_value(self):
        acc = Account.objects.get(id=1)
        bal = acc._meta.get_field('balance').default
        self.assertEqual(bal,0)
    
    def test_unique_name(self):
        acc_type1 = AccountType.objects.get(id=1)
        with self.assertRaises(IntegrityError):
            Account.objects.create(name="Cash",acc_type=acc_type1)
    
    def test_bal_type(self):
        acc=Account.objects.get(id=1)
        bal_type = acc.acc_type.bal_type
        self.assertEqual(bal_type,'D')

    def test_object_name(self):
        acc = Account.objects.get(id=1)
        expected_object_name = f'{acc.name}'
        self.assertEquals(expected_object_name,str(acc))
    
    def test_get_absolute_url(self):
        acc = Account.objects.get(id=1)
        # This will also fail if the urlconf is not defined.
        self.assertEquals(acc.get_absolute_url(), '/fin/acc/1/')

    def test_meta(self):
        acc = Account.objects.get(id=1)
        order = acc._meta.ordering
        self.assertEquals(order,['name'])

    def test_delete_cascade(self):
        AccountType.objects.get(id=1).delete()
        self.assertEqual(len(Account.objects.all()),0)
    
    
class TransactionTest(TestCase):
    def setUp(self):
        acc_type1 = AccountType.objects.create(name = "Assets",bal_type = "D")
        acc_type2 = AccountType.objects.create(name = "Liabilities",bal_type = "C")
        acc1 = Account.objects.create(name="Cash",acc_type=acc_type1)
        acc2 = Account.objects.create(name="Short-Term Debt",acc_type = acc_type2)
        txn1 = Transaction.objects.create(from_acc = acc2, to_acc = acc1, amount = 100, note="Loan from cousin")
        txn2 = Transaction.objects.create(from_acc = acc2, to_acc = acc1, amount = 50, note = 'update', updating = txn1)
    
    def test_date_label(self):
        txn = Transaction.objects.get(id=1)
        field_label = txn._meta.get_field('date').verbose_name
        self.assertEquals(field_label, 'date')
    
    def test_from_acc_label(self):
        txn = Transaction.objects.get(id=1)
        field_label = txn._meta.get_field('from_acc').verbose_name
        self.assertEquals(field_label, 'from account')
    
    def test_to_acc_label(self):
        txn = Transaction.objects.get(id=1)
        field_label = txn._meta.get_field('to_acc').verbose_name
        self.assertEquals(field_label, 'to account')
    
    def test_amount_label(self):
        txn = Transaction.objects.get(id=1)
        field_label = txn._meta.get_field('amount').verbose_name
        self.assertEquals(field_label, 'amount')
    
    def test_note_label(self):
        txn = Transaction.objects.get(id=1)
        field_label = txn._meta.get_field('note').verbose_name
        self.assertEquals(field_label, 'note')
    
    def test_updating_label(self):
        txn = Transaction.objects.get(id=1)
        field_label = txn._meta.get_field('updating').verbose_name
        self.assertEquals(field_label, 'updating')
    
    def test_edited_label(self):
        txn = Transaction.objects.get(id=1)
        field_label = txn._meta.get_field('edited').verbose_name
        self.assertEquals(field_label, 'edited')
    
    def test_note_max_length(self):
        txn = Transaction.objects.get(id=1)
        max_length = txn._meta.get_field('note').max_length
        self.assertEquals(max_length,256)

    def test_object_name(self):
        txn = Transaction.objects.get(id=1)
        expected_object_name = f'TXN on {txn.date} from {txn.from_acc} to {txn.to_acc} for {txn.amount} AED'
        self.assertEquals(expected_object_name,str(txn))
    
    def test_get_absolute_url(self):
        txn = Transaction.objects.get(id=1)
        # This will also fail if the urlconf is not defined.
        self.assertEquals(txn.get_absolute_url(), '/fin/txn/1/')

    def test_meta(self):
        txn = Transaction.objects.get(id=1)
        order = txn._meta.ordering
        self.assertEquals(order,['date'])

    def test_updating(self):
        txn = Transaction.objects.get(id=2)
        prev = txn.updating
        self.assertEquals(prev,Transaction.objects.get(id=1))

    def test_edited_default(self):
        txn = Transaction.objects.get(id=1)
        self.assertEquals(txn.edited,False)

    def test_delete_cascade(self):
        AccountType.objects.get(id=1).delete()
        self.assertEqual(len(Transaction.objects.all()),0)



# Generated by Django 3.0.3 on 2020-03-06 20:32

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('dbaccounting', '0007_auto_20200306_2118'),
    ]

    operations = [
        migrations.AddField(
            model_name='transaction',
            name='updating',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.SET_NULL, to='dbaccounting.Transaction'),
        ),
    ]



from setuptools import setup

# read the contents of the README file
from os import path
this_directory = path.abspath(path.dirname(__file__))
with open(path.join(this_directory, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()
    
setup(
    name='django_dbaccounting',
    long_description=long_description,
    long_description_content_type='text/markdown'
)



# Generated by Django 3.0.3 on 2020-03-04 20:01

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('dbaccounting', '0003_auto_20200304_0000'),
    ]

    operations = [
        migrations.AlterModelOptions(
            name='accounttype',
            options={'ordering': ['-bal_type', 'name']},
        ),
        migrations.AlterModelOptions(
            name='transaction',
            options={'ordering': ['date']},
        ),
        migrations.AlterField(
            model_name='accounttype',
            name='bal_type',
            field=models.CharField(choices=[('C', 'Credit'), ('D', 'Debit')], help_text='Indicate whether the account type is a credit or debit', max_length=1, verbose_name='balance type'),
        ),
    ]



import datetime

from django import forms
from django.forms import ModelForm
from django.core.exceptions import ValidationError
from django.utils.translation import ugettext_lazy as _

from dbaccounting.models import Account, Transaction

class TransactionForm(ModelForm):

    class Meta:
        model = Transaction
        fields = '__all__'
        exclude=['updating','edited']

    def clean(self):
        from_acc_data = self.cleaned_data['from_acc']
        amount_data = self.cleaned_data['amount']
        to_acc_data = self.cleaned_data['to_acc']

        if amount_data<0:
            raise ValidationError(_('Invalid Amount - Must be greater than or equal to 0'))

        if from_acc_data.acc_type.bal_type == 'D':        
            # Check if from_acc has sufficient balance
            if from_acc_data.balance < amount_data:
                raise ValidationError(_('Invalid From Account - Insufficient Funds'))

        if to_acc_data.acc_type.bal_type == 'C':        
            # Check if from_acc has sufficient balance
            if to_acc_data.balance + amount_data > 0:
                raise ValidationError(_('Invalid To Account - Excess Funds'))
        
        ModelForm.clean(self)
    


# Generated by Django 3.0.3 on 2020-03-03 18:45

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    initial = True

    dependencies = [
    ]

    operations = [
        migrations.CreateModel(
            name='Account',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=64, unique=True)),
                ('balance', models.FloatField(default=0, null=True)),
            ],
        ),
        migrations.CreateModel(
            name='AccountType',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=64, unique=True)),
                ('bal_type', models.CharField(choices=[('C', 'Credit'), ('D', 'Debit')], help_text='Indicate whether the account type is a credi or debit', max_length=1)),
            ],
        ),
        migrations.CreateModel(
            name='Transaction',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('date', models.DateTimeField(auto_now_add=True)),
                ('amount', models.FloatField()),
                ('note', models.TextField(blank=True, max_length=256, null=True)),
                ('from_acc', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='from_account', to='dbaccounting.Account')),
                ('to_acc', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='to_account', to='dbaccounting.Account')),
            ],
        ),
        migrations.AddField(
            model_name='account',
            name='acc_type',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='dbaccounting.AccountType'),
        ),
    ]



import datetime

from django.test import TestCase
from django.utils import timezone

from dbaccounting.forms import TransactionForm
from dbaccounting.models import AccountType,Account,Transaction

# Create your tests here.

class TransactionFormTest(TestCase):
    def setUp(self):
        acc_type1 = AccountType.objects.create(name = "Assets",bal_type = "D")
        acc_type2 = AccountType.objects.create(name = "Liabilities",bal_type = "C")
        acc1 = Account.objects.create(name="Short-Term Debt",acc_type=acc_type2)
        acc2 = Account.objects.create(name="Cash",acc_type = acc_type1)
        txn1 = Transaction.objects.create(from_acc = acc1, to_acc = acc2, amount = 100, note="Loan from cousin")    

    # Test Labels
    def test_transaction_from_acc_label(self):
        form = TransactionForm()
        self.assertEquals(form.fields['from_acc'].label, 'From account')

    def test_transaction_to_acc_label(self):
        form = TransactionForm()
        self.assertEquals(form.fields['to_acc'].label, 'To account')

    def test_transaction_amount_label(self):
        form = TransactionForm()
        self.assertEquals(form.fields['amount'].label, 'Amount')
    
    def test_transaction_note_label(self):
        form = TransactionForm()
        self.assertEquals(form.fields['note'].label, 'Note')
    
    # Test Empty Fields
    def test_empty_from_acc(self):
        acc2 = Account.objects.get(id=2)
        with self.assertRaisesMessage(KeyError,'from_acc'):
            form = TransactionForm(data={'to_acc':acc2,'amount':50})
            form.is_valid()
    
    def test_empty_to_acc(self):
        acc1 = Account.objects.get(id=1)
        with self.assertRaisesMessage(KeyError,'to_acc'):
            form = TransactionForm(data={'from_acc':acc1,'amount':50})
            form.is_valid()
    
    def test_empty_amount(self):
        acc1 = Account.objects.get(id=1)
        acc2 = Account.objects.get(id=2)
        with self.assertRaisesMessage(KeyError,'amount'):
            form = TransactionForm(data={'from_acc':acc1,'to_acc':acc2})
            form.is_valid()
    
    # Test Incorrectly Entered Fields
    def test_negative_amount(self):
        acc1 = Account.objects.get(id=1)
        acc2 = Account.objects.get(id=2)
        form = TransactionForm(data={'from_acc':acc1,'to_acc':acc2,'amount':-50})
        self.assertFalse(form.is_valid())
    
    def test_note_over_max_length(self):
        acc1 = Account.objects.get(id=1)
        acc2 = Account.objects.get(id=2)
        note = "test"*300
        form = TransactionForm(data={'from_acc':acc1,'to_acc':acc2,'amount':50,'note':note})
        self.assertFalse(form.is_valid())
    
    # Test a normal form
    def test_transaction_form(self):
        acc1 = Account.objects.get(id=1)
        acc2 = Account.objects.get(id=2)
        form = TransactionForm(data={'from_acc':acc1,'to_acc':acc2,'amount':50,'note':"test"})
        self.assertTrue(form.is_valid())